/* 
 * BSD Telephony Of Mexico "Emiliano" MIPL browser/interpreter, version 1.5  1/1/01
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx 
 *
 *   For documentation on the MIPL language and technology
 *
 *   See http://www.mipl.org
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#include <zap.h>
#include "mipl.h"

#define REC_GOODBYE     "1.vox"               /* thank you for calling, goodbye */
#define REC_ERROR       "2.vox"               /* error er something */
#define REC_INVALID     "3.vox"               /* that was an invalid selection, try again */
#define REC_AUTHREQ     "4.vox"               /* Authorization required */
#define REC_ENTUSER     "5.vox"               /* Please enter ur user # */
#define REC_ENTPSWD     "6.vox"               /* Please enter ur pswd */
#define REC_ENTSINGLE   "7.vox"               /* Please enter 1 single digit */
#define REC_ENTPOUND    "8.vox"               /* Press # after ur entry, or just enter */
#define REC_DIGITS      "9.vox"               /* digits */
#define REC_YESNO       "10.vox"              /* Press 1 for yes, 2 for no */
#define REC_MSGEDTMENU  "11.vox"              /* message edit menu */
#define REC_DISC        "12.vox"              /* message discarded */
#define REC_APPEND      "13.vox"              /* append to your message */ 
#define	REC_XFER	"14.vox"	      /* just a moment please.. */

#define	NUMBERED_EXT ".vox"
#define NUMBER_BASE 1000

#define DAYBASE 1041                /* base of day names */
#define MONTHBASE 1051              /* base of month names */

#define REC_TODAY "1071.vox"         /* today */
#define REC_YESTERDAY "1072.vox"     /* yesterday */
#define REC_AM "1073.vox"            /* AM */
#define REC_PM "1074.vox"            /* PM */
#define REC_OH "1075.vox"            /* Oh */
#define REC_SECOND "1076.vox"        /* second */
#define REC_SECONDS "1077.vox"       /* seconds */
#define REC_AND "1078.vox"			 /* and */

#define LDT_RELDAY      1
#define LDT_DAYOFWEEK   2
#define LDT_DATE        4
#define LDT_YEAR        8
#define LDT_TIME        16
#define LDT_SECS        32
#define LDT_24HR        64

#define	MAXIOT 10000

char *basedir = "/usr/local/mipl";

ZAPIO	iot[MAXIOT];  /* i/o table */
ZAP	*zap;	/* pointer to channel device */

extern int out_mode;
int voicerate;
char didstr[100];
int firsttdd = 0;
int transfermode = 0;
int clidmode = 0;
char *transferfile = "/usr/local/mipl/transfer.default";
float rxgain = 0.0;
float txgain = 0.0;

extern FILE *readurl();
extern char *optarg;
extern int optind;

void oursig(sig)
int sig;
{
        return;
}  

int mipl_flush(int i)
{
int	j;

	  /* clear DTMF buffer */
	if (zap_clrdtmf(zap) == -1)
	   {
		perror("mipl_flush:zap_clrdtmf");
		exit(255);
	   }
	return 0;
}

void miplstr_init()
{
	int i;


	for(i = 0; i < MAXIOT; i++)
	   {
		  /* if no more */
		if (!iot[i].p) break;
		  /* if file, close it */
		if (iot[i].offset >= 0) fclose((FILE *)iot[i].p);
		  /* if a string entry, free it */
		if (iot[i].offset == -2) free((char *)iot[i].p);
		  /* set to 0 */
		iot[i].p = NULL;
	   }
	iot[0].p = NULL;
	return;
}

int miplvstr_url(char *url)
{
	FILE *tfp,*ttfp;
	char a[100],buf[4096];
	int i,n,idx;

	ttfp = readurl(url,0,0);
	if (!ttfp)
	   {
	        fprintf(stderr,"Cannot add URL %s to voice string!!\n",url);
		return(-1);
	   }   
	tfp = tmpfile();
	if (!tfp)
	   {
	   	perror("tmpfile");
	   	exit(255);
	   }
	topcm(ttfp,tfp,voicerate);
	rewind(tfp);
	for(i = 0; i < MAXIOT; i++)
	{
		if (iot[i].p) continue;
		break;
	}
	if (i >= MAXIOT) return(-1);
	iot[i].p = (void *) tfp;
	iot[i].offset = iot[i].length = 0;
	iot[++i].p = NULL;
	return 0;
}

int mipl_vstradd(char *rec)
{
char fname[100],buf[4096];
FILE *fp,*tfp;
int idx,n,i;

	sprintf(fname,"%s/vox%d/%s",basedir,voicerate,rec);
	fp = fopen(fname,"r");
	if (!fp)
	{
		fprintf(stderr,"Cannot open voice file %s",fname);
		return;
	}
	tfp = tmpfile();
	if (!tfp)
	   {
	   	perror("tmpfile");
	   	exit(255);
	   }
	topcm(fp,tfp,voicerate);
	fclose(fp);
	rewind(tfp);
	for(i = 0; i < MAXIOT; i++)
	{
		if (iot[i].p) continue;
		break;
	}
	if (i >= MAXIOT) return(-1);
	iot[i].p = (void *) tfp;
	iot[i].offset = iot[i].length = 0;
	iot[++i].p = NULL;
	return 0;
}

void mipl_ntvstradd(int n)
{
char fname[200];

	sprintf(fname,"%d%s",n,NUMBERED_EXT);
	mipl_vstradd(fname);
}

void vn_do_trip(unsigned int n)
{
int     i,j;

        i = n % 100; /* get remainder after 100 */
        j = n / 100;
        if (j) {
                mipl_ntvstradd(NUMBER_BASE + j); /* Say digit */
                mipl_ntvstradd(NUMBER_BASE + 28); /* Say hundred */
                if (!i) return;
                }
        if (i < 21) mipl_ntvstradd(NUMBER_BASE + i); else {
                mipl_ntvstradd((NUMBER_BASE + 18) + (i / 10)); /* say teen digit */
                if (i % 10) mipl_ntvstradd(NUMBER_BASE + (i % 10)); /* say ones digit */
                }
        return;
}    


int mipl_vnumcat(unsigned int value)
{
long    l,m;

        if (value > 999999999) {
                mipl_ntvstradd(NUMBER_BASE + 31);  /* number too large */
                return(1);
                }
        if (value == 0) {
                mipl_ntvstradd(NUMBER_BASE);
                return(0);
                }
        l = value / 1000000;
        m = value % 1000000;
        if (l) {
                vn_do_trip((int)(l));
                mipl_ntvstradd(NUMBER_BASE + 30); /* put in million */
                }
        if (!m) return(0);
        l = m / 1000; 
        if (l) {
                vn_do_trip((int)(l));
                mipl_ntvstradd(NUMBER_BASE + 29); /* put in thousand */
                }
        if (m % 1000) vn_do_trip((int)(m % 1000));
        return(0);
} 

void vstrliteral(char *str)
{
int     i;

        for(i = 0; str[i]; i++)
           {
                if (str[i] < 32) continue;
                if (str[i] > 126) continue;
                if ((str[i] >= 'a') && (str[i] <= 'z'))
                   {
                        mipl_ntvstradd(str[i] + 136);
                        continue;
                   }
                if (str[i] > 'z')
                         mipl_ntvstradd(str[i] + 142);
                else
                         mipl_ntvstradd(str[i] + 168);
           }
        return;
}


void datecat(time_t t,int flag)
{
struct tm *tm;
time_t	now,midnight;
int	i,dodate = 1;

	  /* get present time */
	time(&now);
	  /* put in tm format */
	tm = localtime(&now);
	  /* calculate value for midnight */
	midnight = now - 
		(tm->tm_sec + (60 * tm->tm_min) + (3600 * tm->tm_hour));
	  /* get the time in tm format */
	tm = localtime(&t);
	if (flag & LDT_RELDAY)
	   {
		if ((t >= midnight) && (t < (midnight + 86400)))
		   {
			mipl_vstradd(REC_TODAY);
			dodate = 0;
		   }
		if ((t < midnight) && (t >= (midnight - 86400)))
		   {
			mipl_vstradd(REC_YESTERDAY);
			dodate = 0;
		   }
	   }
	if ((flag & LDT_DAYOFWEEK) && dodate)
	   {
		mipl_ntvstradd(DAYBASE + tm->tm_wday);
	   }
	if ((flag & LDT_DATE) && dodate)
	   {
		mipl_ntvstradd(MONTHBASE + tm->tm_mon);
		mipl_vnumcat((long) tm->tm_mday);
	   }
	if ((flag & LDT_YEAR) && dodate)
	   {
		mipl_vnumcat(tm->tm_year + 1900);
	   }
	if (flag & LDT_TIME)
	   {
		i = tm->tm_hour;
		if (!(flag & LDT_24HR))
		   {
			if (i == 0) i = 12;
			if (i > 12) i -= 12;
			mipl_vnumcat((long) i);
		   }
		if (i = tm->tm_min)
		   {
			if (i < 10)
			   {
				mipl_vstradd(REC_OH);
			   }
			mipl_vnumcat((long) i);
		   }
		if (!(flag & LDT_24HR))
			mipl_vstradd((tm->tm_hour >= 12) ? REC_PM : REC_AM);
	   }
	if (flag & LDT_SECS)
	   {
		mipl_vstradd(REC_AND);
		mipl_vnumcat((long) tm->tm_sec);
		mipl_vstradd((tm->tm_sec == 1) ? REC_SECOND : REC_SECONDS);
	   }
	return;
}


time_t bldtime(int year,int mon,int day,int hour,int min,int sec)
{

static int montab[]={31,28,31,30,31,30,31,31,30,31,30,31};
time_t	t;
struct tm *tm;
int	i;

	time(&t);
	tm = localtime(&t);
	t = 0;
	if (year > 1970) for(i = 1970; i < year; i++) 
		t += (((i % 4) == 0) ? 31622400 : 31536000) ;
	montab[1] = (((i % 4) == 0) ? 29 : 28);
	if (mon > 1) for(i = 1; i < mon; i++) t += montab[i - 1] * 86400;
	t += ((day - 1) * 86400) + (hour * 3600) + (min * 60) + sec;
	return(t + tm->tm_gmtoff);
}

int miplvstr_cmd(char *cmd,char *value)
{
int     i0,i1,i2,i3;
time_t  t;
struct tm *tm;

    if (!strcmp(cmd,"SRC")) miplvstr_url(value);
	if (!strcmp(cmd,"NUMBER")) mipl_vnumcat(atoi(value));
	if (!strcmp(cmd,"LITERAL")) vstrliteral(value);
    if (!strcmp(cmd,"DATE"))
       {
            if (!value) return(0);
            time(&t);
            ctime(&t);
            sscanf(value,"%d/%d/%d",&i1,&i2,&i3);
            datecat(bldtime(i3,i1,i2,1,0,0),LDT_DATE | LDT_YEAR);
       }
    if (!strcmp(cmd,"TIME"))
       {
            if (!value) return(0);
            time(&t);
            ctime(&t);
            tm = localtime(&t);
            i1 = i2 = i3 = 0;
            i0 = sscanf(value,"%d:%d:%d",&i1,&i2,&i3);
            if (i0 < 3)
              datecat(bldtime(tm->tm_year + 1900,
                tm->tm_mon + 1,tm->tm_mday,i1,i2,0),
                    LDT_TIME);
				            else
              datecat(bldtime(tm->tm_year + 1900,
                tm->tm_mon + 1,tm->tm_mday,i1,i2,i3),
                    LDT_TIME | LDT_SECS);
       }
    return(0);
}
 
static int myrecord(fp,timeout,flag)
FILE *fp;
int timeout;
int flag;
{
ZAPIO	xiot[2];

	xiot[0].p = (void *)fp;
	xiot[0].offset = ftell(fp);
	xiot[0].length = 8 * timeout;
	xiot[1].p = 0;
	return(zap_rec(zap,xiot,((flag) ? ZAP_DTMFINT : 0) | ZAP_HOOKEXIT | ZAP_BEEPTONE));
}


int msg_edit(char *fname,int maxtime,int timeout,int tries)
{
FILE	*fp;
int	i,j,k;
char	buf[10];

	mipl_flush(MIPL_READ);
	fp = fopen(fname,"r");
	if (!fp)
	   {
		fp = fopen(fname,"w+");
		if (!fp)
		   {
			perror("rectemp create");
			exit(255);
		   }
		i = myrecord(fp,maxtime * 1000,1);
		fclose(fp);
		if (i == -2)
		   {
			return(2);
		   }
	   }
	else
	   {
		fclose(fp);
		miplstr_init();
		fp = fopen(fname,"r+");
		if (!fp)
		   {
			perror("rectemp open");
			exit(255);
		   }
		iot[0].p = (void *)fp;
		iot[0].offset = 0;
		iot[0].length = 0;
		iot[1].p = NULL;
		mipl_out(1);
		miplstr_init();
	   }
	fp = fopen(fname,"r+");  /* open it again... */
	if (!fp)
	   {
		perror("rectemp open");
		exit(255);
	   }
	mipl_flush(MIPL_READ);
	for(j = 0; (!tries) || (j < tries); j++)
	   {
		miplstr_init();
		mipl_vstradd(REC_MSGEDTMENU);
		if (!digit_prompt(buf,1,' ',' ',timeout,1,3)) 
		   {
			fclose(fp);
			return(5);
		   }
		switch(buf[0])
		   {
		    case '3': /* discard and start all over */
			fclose(fp);
			fp = fopen(fname,"w+");
			if (!fp)
			   {
				perror("rectemp re-open");
				exit(255);
			   }
			miplstr_init();
			mipl_vstradd(REC_DISC);
			mipl_out(1);
			i = myrecord(fp,maxtime * 1000,1);
			if (i == -2)
			   {
				fclose(fp);
				return(2);
			   }
			ftruncate(fileno(fp),ftell(fp));			
			fclose(fp);
			fp = fopen(fname,"r+");
			if (!fp)
			   {
				perror("rectemp re-open");
				exit(255);
			   }
			j = 0;
			break;
		    case '2': /* append to current recording */
			if (fseek(fp,0L,SEEK_END))
			   {
				perror("fseek");
				exit(255);
			   }
			miplstr_init();
			mipl_vstradd(REC_APPEND);
			mipl_out(1);
			i = myrecord(fp,maxtime * 1000,1);
			if (i == -2)
			   {
				fclose(fp);
				return(2);
			   }
			j = 0;
			if (fseek(fp,0L,SEEK_END))
			   {
				perror("fseek");
				exit(255);
			   }
			ftruncate(fileno(fp),ftell(fp));			
			fclose(fp);
			fp = fopen(fname,"r+");
			if (!fp)
			   {
				perror("rectemp re-open");
				exit(255);
			   }
			j = 0;
			break;
		    case '7' : /* play message as it stands */
			rewind(fp);
			iot[0].p = (void *)fp;
			iot[0].offset = 0;
			iot[0].length = 0;
			iot[1].p = NULL;
			mipl_out(1);
			miplstr_init();
			  /* the mipl_out stuff closes the file handle */
			fp = fopen(fname,"r+");
			if (!fp)
			   {
				perror("rectemp open");
				exit(255);
			   }
			j = 0;
			break;
		    case '9' : /* send file complete */
			fclose(fp);
			return(1);
		    case '0' : /* abort file entirely */
		    case '#' : /* abort file entirely */
			fclose(fp);
			return(0);
		    default:
			miplstr_init();
			mipl_vstradd(REC_INVALID);
			mipl_out(1);
			break;
		   }						
	   }
	fclose(fp);
	return(6);
}

int mipl_edit(edit,fname,maxtime,timeout,tries,tempdir,escurl)
int edit,maxtime,timeout,tries;
char *fname,*tempdir,*escurl;
{
int     r;
char    buf[10],a[300];
FILE	*fp,*fp1;

#ifdef  NORECORD
        return(0);
#else
	mipl_flush(MIPL_READ);
	fp = fopen(fname,"r"); /* open main temp file for read */
	if (!fp)
	   {
		perror("temp file open");
		puts(fname);
		exit(255);
	   }
	sprintf(a,"%sA",fname); /* make secondary file */
	fp1 = fopen(a,"w"); /* open for write */
	if (!fp1)
	   {
		perror("secondary temp file open");
		puts(a);
		exit(255);
	   }
	topcm(fp,fp1,voicerate);
	fclose(fp);
	fclose(fp1);
	if (unlink(fname) == -1) /* delete temp file */
	   {
		perror("temp file unlink");
		exit(255);
	   }
	if (link(a,fname) == -1) /* link to main temp file */
	   {
		perror("temp file link");
		exit(255);
	   }
	if (unlink(a) == -1) /* delete secondary temp file */
	   {
		perror("temp file unlink");
		exit(255);
	   }
        if (!edit)
           {
		fp = fopen(fname,"w");
		if (!fp)
		   {
			perror("rectemp open for write");
			exit(255);
		   }
		r = myrecord(fp,timeout * 1000,1);
                if (r == 1)  /* if tone pressed */
                   {
			   /* if hangup, return as such */
			if (zap_getdtmf(zap,1,NULL,0,0,0,ZAP_HOOKEXIT) == -2)
				r = 2;
                        else if (escurl && zap->dtmfbuf[0]) r = 1;
			zap_clrdtmf(zap);
                   } 
		if (r == -2) r = 2;  /* if a hangup, report as such */
		fclose(fp);
           }
        else
           {
                r = msg_edit(fname,maxtime,timeout,tries);
           }
	if (r == 1) /* if to save recording */
	   {
		fp = fopen(fname,"r"); /* open main temp file for read */
		if (!fp)
		   {
			perror("temp file open");
			puts(fname);
			exit(255);
		   }
		sprintf(a,"%sA",fname); /* make secondary file */
		fp1 = fopen(a,"w"); /* open for write */
		if (!fp1)
		   {
			perror("secondary temp file open");
			puts(a);
			exit(255);
		   }
		frompcm(fp,fp1,voicerate);
		fclose(fp);
		fclose(fp1);
		if (unlink(fname) == -1) /* delete temp file */
		   {
			perror("temp file unlink");
			exit(255);
		   }
		if (link(a,fname) == -1) /* link to main temp file */
		   {
			perror("temp file link");
			exit(255);
		   }
		if (unlink(a) == -1) /* delete secondary temp file */
		   {
			perror("temp file unlink");
			exit(255);
		   }
	   }
	mipl_flush(MIPL_READ);
        return(r);
#endif
}   

int mipltext_add(char *str)
{
int	i;
	
	for(i = 0; i < MAXIOT; i++)
	{
		if (iot[i].p) continue;
		break;
	}
	if (i >= MAXIOT) return(-1);
	iot[i].p = (void *) malloc(strlen(str) + 1);
	if (iot[i].p == NULL)
	   {
		perror("malloc");
		exit(255);
	   }
	sprintf(iot[i].p,"%s\n",str);
	iot[i].offset = -2;  /* indicate a string */
	iot[i].length = 0;
	iot[++i].p = NULL;
	return 0;
}

int mipl_askauth(char *realm,char *user,char *pswd)
{
char a[300];
int	t,v;

	if (out_mode == MIPL_TEXTONLY) /* if in TDD mode */
	   {
		if (zap_sendtdd(zap,"Authentication is Required.\n") < 0) exit(0);
		sprintf(a,"Please enter user number for Authentication realm \"%s\", followed by the return key.\n",realm);
		if (zap_sendtdd(zap,a) < 0) exit(0);
		if (!ttget(user,20,1,30)) return(0);
		if (zap_sendtdd(zap,"Please enter your password, followed by the return key.\n") < 0) exit(0);
		if (!ttget(pswd,20,1,30)) return(0);
		return(1);
	   }
	miplstr_init();
	sprintf(a,"Authentication is required for realm -- %s.\n",realm);
	mipl_vstradd(REC_AUTHREQ); 
	mipl_vstradd(REC_ENTUSER); 
	if (!mipl_prompt(user,30,1,20,0)) return(0);
	miplstr_init();
	mipl_vstradd(REC_ENTPSWD);   
	if (!mipl_prompt(pswd,30,1,20,0)) return(0);
	return(1);
}

int mipl_error(char* str)
{
	printf("%% %s\n",str);
	miplstr_init();
	mipl_vstradd(REC_ERROR);
	mipl_out(1);
	return 0;
}

mipl_voicerate(rate)
int rate;
{
	voicerate = rate;
	return;
}

mipl_out(intflg)
int intflg;
{
int	i;

	if (out_mode == MIPL_TEXTONLY)  /* if in TDD mode */
	   {
		for(i = 0; i < MAXIOT; i++)
		   {
			if (!iot[i].p) break;  /* if last one */
			   /* ignore if not text */
			if (iot[i].offset != -2) continue;
			if (zap_sendtdd(zap,(char *)iot[i].p) < 0) exit(0);
		   }
		return(0); /* TDD is half duplex, cant interrupt */
	   }
	if (iot[0].p) i = mipl_play(intflg); else i = 0;
	miplstr_init();
	return(i);
}

ttget(instr,max,intflag,to)
char *instr;
int max; /* max chars */
int intflag; /* interrupt flag */
int to; /* timeout */
{
int	c,i,j,k;
time_t	t;

	usleep(100000);  /* wait for end of echo and tx to stop */
	time(&t);
	if (to)
	   {
		signal(SIGALRM,oursig);
		siginterrupt(SIGALRM,1);
		sigsetmask(0);
		alarm(to);
	   }
	k = 0;
	for(;;)
	   {
		i = zap_gettdd(zap);
		if (i == -2) exit(0);
		alarm(0L);
		signal(SIGALRM,SIG_DFL);
		if (i == -1) 
		   {
			if (errno == EINTR)
			   {
				instr[k] = 0;
				return(0);
			   }
			perror("tdd read");
			exit(255);
		   }
		c = i & 0x7f;
		if (i < 1) continue;
		if (c == '\r') /* end of input */
		   {
			instr[k] = 0;
			sleep(1);
			if (zap_sendtdd(zap,"\n") < 0) exit(0);
			return(1);
		   }
		if (c == '\n') continue;
		if (c == 8) /* erase */
		   {
			sleep(1);
			if (zap_sendtdd(zap,"   INPUT CANCELLED.  GA   \n") < 0) exit(0);
			instr[k = 0] = 0;
			continue;
		   }
		instr[k++] = c;
		if (k >= max)
		   {
			instr[k] = 0;
			sleep(1);
			if (zap_sendtdd(zap,"\n") < 0) exit(0);
			return(1);
		   }
	   }
}

int mipl_prompt(buf,max,intflag,timeout,mode)
char *buf;
int max,intflag,timeout,mode;
{
char    a[100];

	   /* if in TDD mode */
	if (out_mode == MIPL_TEXTONLY)
	   {
		if (mode == MIPL_ASK)
		   {
			mipltext_add("Enter Y for yes, N for no.  GA  \n");
		   }
		else
		   {
	                if (max > 1)
	                   {
	                        mipltext_add("Hit RETURN after your entry,");
				sprintf(a," or just enter %d characters.  GA  \n",max);
	                        mipltext_add(a);
	                   } else mipltext_add("Enter a single character.  GA  \n"); 
		   }
		mipl_out(intflag);
		return(ttget(buf,max,intflag,timeout));
	   }
        if (mode == MIPL_DATAIN)
           {
	        if (max < 2) mipl_vstradd(REC_ENTSINGLE);
                else mipl_vstradd(REC_ENTPOUND);
           }
        if (mode == MIPL_ASK)
           {
                mipl_vstradd(REC_YESNO);
           }
        if (!digit_prompt(buf,((max > 1) ? 30 : 1),((max > 1) ? '*' : ' '),'#',
                timeout * 10,1,((intflag) ? 3 : 1)))
           {
		miplstr_init();
                return(0);
           }
	miplstr_init();
        if (!buf[0]) return(1);
        if (buf[strlen(buf) - 1] == '#')
                buf[strlen(buf) - 1] = 0;
	buf[max] = 0;
        return(1);
}

int digit_prompt(buf,digits,erase,term,time,retries,flag)
char *buf;  /* touchtone buffer */
int digits;  /* number of digits */
char erase;  /* erase digit */
char term;  /* terminate digit */
int time;  /* timeout time */
int retries;  /* number of times to try */
int flag;  /* flag value */
{
          /* if prompt first */
        if (flag & 1)
           {
                  /* play the prompt */
                if (iot[0].p) mipl_play((flag & 2) != 0);
           }
          /* keep trying to get tones */
        while(retries--)
           {
                if (!(flag & 2)) mipl_flush(MIPL_READ);
                if (mipl_input(buf,digits,erase,term,time)) return(1);
                if (retries && iot[0].p) mipl_play((flag & 2) != 0);
           }
        return(0);
}  

int mipl_play(intflg)
int intflg;
{
int	i;

	  /* play the stuff */
	i = zap_play(zap,iot,((intflg) ? ZAP_DTMFINT : 0) | ZAP_HOOKEXIT);
	if (i == -1) /* if was error, report as such and exit */
	   {
		perror("zap_play");
		exit(255);
	   }
	  /* if hung up, exit */
	if (i == -2) exit(0);
	return(i);
}

int mipl_input(buf,digits,erase,term,time)
char *buf;
int digits,time;
char erase,term;
{
int	i;
char	a[2];

	buf[0] = 0;  /* terminate buffer */
	  /* set up the termintion string */
	a[0] = term;
	a[1] = 0;
	  /* get the DTMF digits */
	i = zap_getdtmf(zap,digits,a,erase,0,time * 100,ZAP_HOOKEXIT);
	if (i == -1) /* if was error, report as such and exit */
	   {
		perror("zap_play");
		exit(255);
	   }
	  /* if hung up, exit */
	if (i == -2) exit(0);
	strcpy(buf,zap->dtmfbuf);
	zap_clrdtmf(zap);  /* clear the buffer */
	return(i);
}
	

void main(argc,argv)
int argc;
char *argv[];
{
int i,ndid;
char	*cp,a[1000],b[200],c,name[256],number[256];


        printf("Emiliano MIPL, Version %s\n",MIPL_VERSION);
        printf("Zapata, BSD open-source telephony voice-based MIPL Browser\n\n");
	printf("The technologies, software, hardware, designs, drawings,\n");
	printf("schematics, board layouts and/or artwork, concepts, methodologies\n");
	printf("(including the use of all of these, and that which is derived\n");
	printf("from the use of all of these), all other intellectual properties\n");
	printf("contained herein, and all intellectual property rights have been\n");
	printf("and shall continue to be expressly for the benefit of all mankind,\n");
	printf("and are perpetually placed in the public domain, and may be used,\n");
	printf("copied, and/or modified by anyone, in any manner, for any legal\n");
	printf("purpose, without restriction.\n\n");


	  /* catch termination signals */
	signal(SIGINT,exit);
	signal(SIGHUP,exit);
	signal(SIGTERM,exit);
	signal(SIGQUIT,exit);


	  /* parse options */
	while((c = getopt(argc,argv,"+cxXf:r:t:")) != -1)
	   {
		switch(c)
		   {
		    case 'f':  /* transfer file */
			transferfile = optarg;
			break;
		    case 'X':  /* non-centrex transfer */
			transfermode = 2;
			break;
		    case 'x':  /* centrex transfer */
			transfermode = 1;
			break;
		    case 'c' : /* Caller*ID mode */
			clidmode = 1;
			break;
		    case 'r': /* rx gain in db */
			if (sscanf(optarg,"%f",&rxgain) != 1)
			   {
				fprintf(stderr,"Invalid rx gain specification\n");
				exit(255);
			   }
			break;
		    case 't': /* send gain in db */
			if (sscanf(optarg,"%f",&txgain) != 1)
			   {
				fprintf(stderr,"Invalid tx gain specification\n");
				exit(255);
			   }
			break;
		   }
	   }
	argc -= (optind - 1);
	argv += (optind - 1);

	  /* if insufficient args, report as such */
	if (argc < 3)
	   {
		puts("Usage... emiliano [-xX filename] [-r gain] [-t gain] channel_device #did/rings [Start_URL]");
		exit(1);
	   }

	  /* attempt to parse number of did/rings */
	if (sscanf(argv[2],"%d",&ndid) != 1)
	   { /* if error, report and exit */
		fprintf(stderr,"Invalid # of did/rings specification\n");
		exit(255);
	   }

	  /* set up default URL */
	cp = "http://www.lambdatel.com/mipl";
	  /* point to specified URL, if any */
	if (argc > 3) cp = argv[3];

	zap = zap_open(argv[1]); /* attempt to open device channel */
	if (!zap) /* if error, report and exit */
	   {
		perror("device open");
		exit(255);
	   }

	memset(iot,0,sizeof(iot));  /* clear iot structure */

	didstr[0] = 0; /* initialize did digit string */
	number[0] = 0; /* initialize number string */
	if (ndid < 0)  /* if not E&M */
	   {
		if (clidmode) /* if to receive caller*id */
		   {
			i = zap_clid(zap,number,name);
			if (i == -2) exit(0); /* if hung up */
			if (i == -1)
			   {
				perror("zap_clid");
				exit(255);
			   }
			printf("Number: %s, Name: %s\n",number,name);
			  /* if just waiting for 1 ring, answer now */
			if (ndid == -1)
			   {
				if (zap_sethook(zap,ZAP_OFFHOOK) == -1)
				   {
					perror("zap_hook");
					exit(255);
				   }
			   }
			else /* otherwise, wait for the rest of them */
			   {
				ndid++; /* make count less one (negatively) */
				  /* wait for additional rings, and answer it */
			        if (zap_waitcall(zap,-ndid,ZAP_OFFHOOK,6) == -1)
			           {
					  /* if timeout, exit */
					if (errno == EINTR) exit(0);
			                perror("zap_waitcall");
			                exit(255);
			           }
			   }
		   }
		else /* is non-caller*id, so do it normally */
		   {
			  /* wait for call, and answer it */
		        if (zap_waitcall(zap,-ndid,ZAP_OFFHOOK,
				((ndid != -1) ? 6 : 0)) == -1)
		           {
				  /* if timeout, exit */
				if (errno == EINTR) exit(0);
		                perror("zap_waitcall");
		                exit(255);
		           }
		   }
		usleep(500000); /* wait 1/2 sec */
	   }
	else /* is E&M */
	   {
		  /* wait for incomming seizure */
	        if (zap_waitcall(zap,1,ZAP_ONHOOK,0) == -1)
	           {
	                perror("zap_waitcall");
	                exit(255);
	           }
	        zap_clrdtmf(zap); /* clear DTMF buffer */
		if (clidmode)
		   {
		        printf("winking on %s\n",argv[1]);
		        zap_wink(zap); /* wink at trunk */
			  /* get initial '*' */
		        if (!zap_getdtmf(zap,10,"*",' ',0,5000,ZAP_HOOKEXIT))
			   {
				if (i == -1)
				   {
					perror("zap_getdtmf");
					exit(0);
				   }
				if (i == -2) exit(0);
				fprintf(stderr,"Timeout getting LEAD_IN digit on %s\n",argv[1]);
				exit(0);
			   }
			zap->dtmfn = 0;
			  /* get ANI digits */
		        if (!zap_getdtmf(zap,15,"*",' ',0,5000,ZAP_HOOKEXIT))
			   {
				if (i == -1)
				   {
					perror("zap_getdtmf");
					exit(0);
				   }
				if (i == -2) exit(0);
				fprintf(stderr,"Timeout getting ANI digits on %s\n",argv[1]);
				exit(0);
			   }
			number[0] = 0;
			if (zap->dtmfbuf[0]) zap->dtmfbuf[strlen(zap->dtmfbuf) - 1] = 0;
			if (zap->dtmfbuf[0]) strcpy(number,zap->dtmfbuf);
			zap->dtmfn = 0;
			  /* get DNIS digits */
		        if (!zap_getdtmf(zap,20,"*",' ',0,5000,ZAP_HOOKEXIT))
			   {
				if (i == -1)
				   {
					perror("zap_getdtmf");
					exit(0);
				   }
				if (i == -2) exit(0);
				fprintf(stderr,"Timeout getting DNIS digits on %s\n",argv[1]);
				exit(0);
			   }
			didstr[0] = 0;
			if (zap->dtmfbuf[0]) zap->dtmfbuf[strlen(zap->dtmfbuf) - 1] = 0;
			if (zap->dtmfbuf[0]) strcpy(didstr,zap->dtmfbuf);
			printf("Got ANI: %s, DNIS: %s on %s\n",number,didstr,argv[1]);
		   }
		else
		   {
			if (ndid) /* if DID digits to get */
			   { 
				printf("winking on %s\n",argv[1]);
			        zap_wink(zap); /* wink at trunk */
				  /* get DID digits */
			        if (zap_getdtmf(zap,ndid,NULL,' ',0,5000,ZAP_HOOKEXIT) != ndid)
				   {
					fprintf(stderr,"Error getting DID digits on %s\n",argv[1]);
					exit(0);
				   }
				  /* save DID digits */
				strcpy(didstr,zap->dtmfbuf);
				printf("Got DID: %s on %s\n",didstr,argv[1]);
			   }
		   }
		  /* clear DTMF buffer */
		zap_clrdtmf(zap);
		  /* take trunk off hook (no salad bar, please!!) */
		if (zap_sethook(zap,ZAP_OFFHOOK) == -1)
		   {  /* if error, report it and exit */
			perror("zap_offhook");
			exit(255);
		   }
	   }
	if (zap_clearlinks(zap) == -1)
	   {
		perror("zap_clearlinks");
		exit(255);
	   }
	/* set bridge 2 listening to bridge 1 */
	if (zap_makelink(zap,1,2) == -1)
	   {
		perror("zap_makelink");
		exit(255);
	   }
	out_mode = MIPL_VOICEONLY;	
	voicerate = 8000;
	if (clidmode && number[0] && (number[0] != 'P') && (number[0] != '0'))
	   {
		miplstr_init();
		for(i = 0; number[i]; i++)
		   {
			mipl_vnumcat(number[i] - '0');
		   }
		mipl_play(1);
	   }
	  /* create URL using URL specification as template */
	sprintf(a,cp,didstr);
	  /* make unique identifier for temp file naming purposes */
	sprintf(b,"EMIL%s",argv[1]);
	  /* remove slashes from the name */
	for(i = 0; b[i]; i++) if (b[i] == '/') b[i] = '_';
	miplmain(a,b);
	miplstr_init();
	mipl_vstradd(REC_GOODBYE);
	mipl_play(1);
	exit(0);
}

/* do the conference demo */
conftest(url)
char *url;
{
int	i,flag;

	if (sscanf(url,"%i",&flag) != 1)
	   {
		fprintf(stderr,"Invalid conf specified: %s\n",url);
		exit(255);
	   }
	  /* set gains */
	if (zap_gains(zap,rxgain,txgain) == -1)
	   {
		perror("zap_gains");
		exit(255);
	   }
	voicerate = 8000;
	  /* play greeting */
	i = zap_playf(zap,"/usr/local/mipl/conf.pcm",ZAP_HOOKEXIT);
	if (i == -1) /* if was error, report as such and exit */
	   {
		perror("zap_play");
		exit(255);
	   }
	  /* if hung up, exit */
	if (i == -2) exit(0);
	  /* set to announce mode on bridge */
	if (zap_conf(zap,0,flag & 0xff,
		ZAP_CONFANN | ZAP_CONF_TALKER | ZAP_CONF_LISTENER) == -1)
	   {
		perror("zap_conf");
		exit(255);
	   }
	  /* play entering announcement on the bridge */
	i = zap_playf(zap,"/usr/local/mipl/5110.pcm",ZAP_HOOKEXIT);
	if (i != -2)  /* if wasnt a hangup */
	   {
		  /* go to normal conf mode */
		if (zap_conf(zap,0,flag & 0xff,
			ZAP_CONF | ZAP_CONF_TALKER | ZAP_CONF_LISTENER) == -1)
		   {	
			perror("zap_conf");
			exit(255);
		   }
		  /* wait for hangup (er something) */
		if (zap_waitevent(zap) == -1)
		   {
			perror("zap_waitevent");
			exit(255);
		  }
	   }
	  /* set back to announcement mode */
	if (zap_conf(zap,0,flag & 0xff,
		ZAP_CONFANN | ZAP_CONF_TALKER | ZAP_CONF_LISTENER) == -1)
	   {
		perror("zap_conf");
		exit(255);
	   }
	   /* play leaving announcement to the bridge */
	zap_playf(zap,"/usr/local/mipl/5111.pcm",0);
	exit(0);
}

mipl_xfer(url)
char *url;
{
int	i;
FILE	*fp;
char	a[100];

	  /* if we cant transfer, error */
	if (!transfermode)
	   {
		mipl_error("Transfer Not Supported");
		return;
	   }
/* The transfer file contains substrings (first part of numbers) that are
valid to be dialed, one on each line. Blank lines, and comments starting
with '#' are okay too */
	  /* open the transfer file */
	fp = fopen(transferfile,"r");
	if (!fp)
	   {
		perror("transfer file open");
		puts(transferfile);
		exit(255);
	   }
	i = 0;  /* clear found flag */
	/* we need to find a match to be okay */
	while(fgets(a,sizeof(a) - 1,fp))
	   {
		  /* skip completely blank lines */
		if (!a[0]) continue;
		  /* strip lf */
		if (a[strlen(a) - 1] == '\n') a[strlen(a) - 1] = 0;
		  /* skip blank lines */
		if (!a[0]) continue;
		  /* skip comments */
		if (a[0] == '#') continue;
		if (!strncmp(url,a,strlen(a)))
		   {
			i = 1;  /* set found flag */
			break;
		   }
	   }
	fclose(fp);
	if (!i)  /* if not found */
	   {	
		sprintf(a,"Invalid Destination number: %s",url);
		mipl_error(a);
		return;
	   }
	miplstr_init();
	mipl_vstradd(REC_XFER);
	mipl_out(1);
	  /* flush events (line transitions) */
	if (zap_flushevent(zap) == -1)
	   {
		perror("zap_flushevent");
		exit(255);
	   }
	  /* flash hookswitch */
	if (zap_flash(zap) == -1)
	   {
		perror("zap_flash");
		exit(255);
	   }
	usleep(1000000); /* wait 1 second */
	  /* flush events (line transitions) */
	if (zap_flushevent(zap) == -1)
	   {
		perror("zap_flushevent");
		exit(255);
	   }
	  /* dial it */
	if (zap_dial(zap,url,0) == -1)
	   {
		  /* on a 1AESS, yes one of the last of its kind, you can get
		    a line transition in the MIDDLE of the last digit, which
		    is okay, since the CO obviously has the digit already,
		    because otherwise we wouldnt get the transition */
		if (errno != ELAST) /* if not an event */
		   {
			perror("zap_dial");
			exit(255);
		  }
	   }
	usleep(750000); /* wait 3/4 second */
	  /* flash hookswitch */
	if (zap_flash(zap) == -1)
	   {
		perror("zap_flash");
		exit(255);
	   }
	  /* flush events (line transitions) */
	if (zap_flushevent(zap) == -1)
	   {
		perror("zap_flushevent");
		exit(255);
	   }
	if (transfermode == 2) /* if not centrex */
	   {
		/* since we arent a centrex line, and are using
		     regular 3-way calling, hang out until line transition */
		if (zap_waitevent(zap) == -1)
		   {
			perror("zap_waitevent");
			exit(255);
		   }
	   }
	else /* if centrex */
	   {
		/* wait 1 second after dropping, thus transferring */
		usleep(1000000); /* wait 1 second */
	   }
	return;
}
