/* 
 * Mu law 8k bit pcm to Oki-style ADPCM converter, part of:
 *  BSD Telephony Of Mexico "Emiliano" MIPL browser/interpreter, version 1.5  1/1/01
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx 
 *
 *   For documentation on the MIPL language and technology
 *
 *   See http://www.mipl.org
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction. 
 *
 */

/* standard headers */
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
/* step size index shift table */
static short indsft[8]={-1, -1, -1, -1, 2, 4, 6, 8};
/* step size table
    stpsz[i]=floor[16*(11/10)^i] */
static short stpsz[49]=
{16,17,19,21,23,25,28,31,34,37,41,45,50,55,60,66,73,80,88,
97,107,118,130,143,157,173,190,209,230,253,279,307,337,371,
408,449,494,544,598,658,724,796,876,963,1060,1166,1282,1411,
1552};
/* nibble to bit map */
static short nbl2bit[16][4]={
{1,0,0,0},{1,0,0,1},{1,0,1,0},{1,0,1,1},{1,1,0,0},{1,1,0,1},
{1,1,1,0},{1,1,1,1},{-1,0,0,0},{-1,0,0,1},{-1,0,1,0},{-1,0,1,1},
{-1,1,0,0},{-1,1,0,1},{-1,1,1,0},{-1,1,1,1}};
/* step size index */
static short ssindex=0;
/* the current pcm signal */
static short signal=0;
/* length of the current read buffer */
static long length;
/* offset into the current read buffer */
static long offset;
/* nibble of the byte */
static short nibble;
/* read buffer */
static unsigned char readbuf[8192];
static short linbuf[8192];
/* write buffer */
static unsigned char writebuf[8192];
/* wave file format */
static short format;
static short decode(unsigned char);
unsigned char adpcm(short);
extern int resamplemode;

int frompcm(ifp,ofp,rate)
FILE *ifp,*ofp;
int rate;
{
int	length;
int	ifh,ofh;
FILE	*tfp,*tfpi;
short	lin;
extern short	Imp[],ImpD[];
extern unsigned short LpScl, Nwing, Nmult;
extern int	fullmulaw[];
extern FILE *fin,*fout;
unsigned char ensig;

    ifh = fileno(ifp);
    ofh = fileno(ofp);


	/* downsample from 8000 hz to 6000 hz */
    if (rate != 8000)
      {
	tfp = tmpfile();
	if (!tfp)
	   {
		perror("tmpfile");
		exit(255);
	   }
	tfpi = tmpfile();
	if (!tfpi)
	   {
		perror("tmpfile1");
		exit(255);
	   }
	length=8192;
	while(length == 8192)
	   {
		length=fread(readbuf,1,8192,ifp);
	        for(offset=0;offset<length;++offset)
		   {
			linbuf[offset] = fullmulaw[readbuf[offset]];
		   }
		fwrite(linbuf,2,length,tfpi);
	   }
	rewind(tfpi);
	fin = tfpi;
	fout = tfp;
	resamplemode = 0;
	resample(Imp, ImpD, LpScl, Nmult, Nwing, .75662424, 0);
	rewind(tfp);
	close(tfpi);
      }
    length=8192;
    while(length==8192){
	if (rate == 8000)
		length=fread(readbuf,1,8192,ifp);
	else
		length=fread(linbuf,2,8192,tfp);
        for(offset=0;offset<length;++offset){
	    if (rate == 8000)
		lin = fullmulaw[readbuf[offset]];
	    else
		lin = linbuf[offset];
            ensig=(unsigned char)(adpcm(lin/3)*16);
            ++offset;
	    if (rate == 8000)
		lin = fullmulaw[readbuf[offset]];
	    else
		lin = linbuf[offset];
            ensig+=adpcm(lin/3);
            writebuf[offset/2]=ensig;
 	    }
	    write(ofh,writebuf,(short)(length/2));
    }
    if (rate != 8000) fclose(tfp);
    return(0);
}
/****************************************************************
*	 NAME : decode(encoded)
* DESCRIPTION : does the actual adpcm decode
*	INPUT : the encoded nibble from the adpcm file
*      OUTPUT : the index into the step size table for the next encode
*     RETURNS : the encoded difference
****************************************************************/
static short decode(unsigned char encoded)
{
    short diff,step;
    step=stpsz[ssindex];
    diff=nbl2bit[encoded][0]*(
	step*nbl2bit[encoded][1]+
	(step/2)*nbl2bit[encoded][2]+
	(step/4)*nbl2bit[encoded][3]+
	(step/8));
    ssindex=ssindex+indsft[(encoded%8)];
    if(ssindex<0)
	ssindex=0;
    if(ssindex>48)
	ssindex=48;
    return(diff);
}
unsigned char adpcm(short csig)
{
    short diff,step;
    unsigned char encoded;
    step=stpsz[ssindex];
    if(csig>2047){
	csig=2047;
    }
    if(csig<-2047){
	csig=-2047;
    }
    diff=csig-signal;
    if(diff<0){
	encoded=8;
	diff=-diff;
    }
    else
	encoded=0;
    if(diff>=step){
	encoded+=4;
	diff-=step;
    }
    step/=2;
    if(diff>=step){
	encoded+=2;
	diff-=step;
    }
    step/=2;
    if(diff>=step)
	encoded+=1;
    signal+=decode(encoded);
    return(encoded);
}
