/* 
 * Master controlling process, part of:
 *  BSD Telephony Of Mexico "Emiliano" MIPL browser/interpreter, version 1.5  1/1/01
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx 
 *
 *   For documentation on the MIPL language and technology
 *
 *   See http://www.mipl.org
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction. 
 *
 */

#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>

static int portpid[70] = {0} ;
static char portcmd[70][128] = {0} ;
static int shutting = 0;
char	space[128 * 8];

int findport(pid)
int pid;
{
int	i;

	for(i = 1; i < 70; i++) if (portpid[i] == pid) return(i);
	return(0);
}

void ourhandle(sig)
int sig;
{
int	cstat,hispid,hisport,i;

	if ((sig == SIGTERM) || (sig == SIGINT) || (sig == SIGQUIT) ||
	  (sig == SIGHUP))
	   {
		orderly();
		exit(0);
	   }
	if (shutting) return;
	if (sig == SIGCHLD)
	   {
		  /* get the status of the death */
		hispid = wait(&cstat);
		hisport = findport(hispid);
		if (cstat && (cstat != SIGTERM) && (cstat != SIGINT))
		   {
			fprintf(stderr,"master:GASP!! -- Child PID %d (port %d) died with status %x\n",
				hispid,hisport,cstat & 0xffff);
			orderly();
			exit(255);
		   }
		if (hisport) portpid[hisport] = 0;
		signal(SIGCHLD,ourhandle);
	   }
	return;
}

orderly()
{
int	i;

	shutting = 1;
	fflush(stderr);
	fprintf(stderr,"Doing orderly shutdown\n");
	for(i = 0; i < 70; i++) 
		if (portpid[i]) kill(portpid[i],SIGTERM);
	return;
}

main(argc,argv)
int argc;
char *argv[];
{
int	i,j,k;
FILE	*fp;
char	a[128],cmd[128],*b[8];
time_t	t;


	fp = fopen("/usr/local/mipl/config.mipl","r");
	if (!fp)
	   {
		perror("master:Config file open");
		exit(255);
	   }
	while(fgets(a,127,fp) != NULL)
	   {
		if (!a[0]) continue;
		if (a[strlen(a) - 1] == '\n') a[strlen(a) - 1] = 0;
		if (!a[0]) continue;
		if (a[0] == '#') continue;
		if (sscanf(a,"%d %n",&i,&j) != 1) 
		   {
			fprintf(stderr,"master:Invalid config line %s\n",a);
			continue;
		   }
		if ((i < 1) || (i >= 70))
		   {
			fprintf(stderr,"master:Port number %d is invalid\n",i);
			continue;
		   }
		strcpy(&portcmd[i][0],&a[j]);
	   }
	fclose(fp);

	signal(SIGCHLD,ourhandle);
	signal(SIGTERM,ourhandle);
	signal(SIGQUIT,ourhandle);
	signal(SIGHUP,ourhandle);
	signal(SIGINT,ourhandle);

	time(&t);
	printf("System starting at %s\n",ctime(&t));

	for(;;)
	   {
		for(i = 1; i < 70; i++)
		   {
			if (portcmd[i][0] && (!portpid[i]))
			   {
				for(j = 0; j < 8; j++) b[j] = &space[j * 128];
				j = sscanf(&portcmd[i][0],
				    "%s %s %s %s %s %s %s %s",
					b[0],b[1],b[2],b[3],b[4],
					    b[5],b[6],b[7]);
				b[8] = b[j] = NULL;
				portpid[i] = k = fork();
				if (k == -1)
				   {
					fprintf(stderr,"master:Error forking!!!\n");
					exit(255);
				   }
				if (k == 0)
				   {
					if (execv(b[0],b) == -1)
					   {
						perror("master:exec");
						exit(255);
					   }
					exit(255);
				   }
			   }
		   }
		pause();
	   }
}

