/* 
 * MIPL-related Input/Output routines, part of:
 *  BSD Telephony Of Mexico "Emiliano" MIPL browser/interpreter, version 1.5  1/1/01
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx 
 *
 *   For documentation on the MIPL language and technology
 *
 *   See http://www.mipl.org
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction. 
 *
 */

/* #define	DUMPRX  */
/* #define	CACHEDEBUG  */
/* #define	SSLDEBUG */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <netdb.h>
#include <ctype.h>
#ifdef	ISC
#include <malloc.h>
#endif
#include "mipl.h"

#ifdef	MIPL_SSL
#include <errno.h>
#include <memory.h>
#include <arpa/inet.h>
#include <openssl/rsa.h>       /* SSLeay stuff */
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif 

#ifdef MIPL_CACHE
#include <gdbm.h>

#define CACHETRIES 20
#define CACHEDIR "/usr/local/miplcache"
#define CACHEDB "/usr/local/miplcache/cache"
#define CACHETMP "MIPLXXXXXX"
#endif

#ifdef	MIPL_SSL
#define CHK_NULL(x) if ((x)==NULL) exit (1)
#define CHK_ERR(err,s) if ((err)==-1) { perror(s); exit(1); }
#define CHK_SSL(err) if ((err)==-1) { ERR_print_errors_fp(stderr); exit(2); } 
#endif


#ifndef	NORECORD
extern char rectemp[];
#endif
#ifdef	MIPL_CACHE
int mainchanged = 0;
#endif
char ourlabel[MAXURL]; /* our label n stuff */
char defstuff[MAXURL]; /* default URL prefix */
extern AUTH auth;
extern TAG maintag;
extern FORM form;
extern char outbuf[];
char *myhostname; /* pointer to my host name, if any */
char bigbuf[1024];
int bign = 0;
int bigindex = 0;
FILE *fp = 0;        
FILE dummyf;
int usessl;
#ifdef	MIPL_SSL
SSL_METHOD *meth;
SSL_CTX* ctx;  
SSL*     ssl;  
#endif

miplmain(url,str)
char *url;
char *str;
{
int c,i,err;
char buf[200];
struct hostent *host;
long	addr;
#ifdef	MIPL_CACHE
GDBM_FILE	dbf;

	mkdir(CACHEDIR,0777);
        dbf = gdbm_open(CACHEDB,0,GDBM_READER,0,NULL);
        if (!dbf)
           {
                dbf = gdbm_open(CACHEDB,0,GDBM_NEWDB,0666,NULL);
                if (!dbf)
                   {
                        printf("Cannot create CACHE db file!!!\n");
                        exit(255);
                   }
           }
        gdbm_close(dbf);  
#endif

#ifdef	MIPL_SSL
	SSLeay_add_ssl_algorithms();
	meth = SSLv2_client_method();
	SSL_load_error_strings();
	ctx = SSL_CTX_new (meth);
	CHK_NULL(ctx);
	CHK_SSL(err);  
#endif
	gethostname(buf,80);
	host = gethostbyname(buf);
	if (!host)
	   {
		perror("Gethostbyname local");
		exit(255);
	   }
	myhostname = 0;
	if (host->h_name)
	   {
		myhostname = malloc(strlen(host->h_name) + 1);
		if (!myhostname)
		   {
			perror("malloc");
			exit(255);
		   }
		strcpy(myhostname,host->h_name);
	   }
	strcpy(defstuff,"file://");

#ifndef	NORECORD
        mkdir(RECTEMP,0755);
        sprintf(rectemp,"%s/%s",RECTEMP,str);
        unlink(rectemp); 
#endif
	if (openurl(url,0))
	   {
		mipl_error("Cant open URL!!!");
		return;
	   }
	while(runscript(ourlabel) == 1);
#ifndef	NORECORD
	unlink(rectemp);
#endif
	return;
}

FILE *readurl(url,method,mainflag)
char *url;
int method;
int mainflag;
{
char	*cp,*cpx,*cpy,*cp1,*cp2,*cp3,*cp4,*cp5,a[MAXURL*2 + 10],b[MAXURL*2];
char	buf[512],*strs[100],*strs1[100],cmod[100],cname[512],*cpt,a1[512];
int	c,cl,i,j,k,s,clen,nocache,cachehit,cache_read;
struct hostent *host;
long	addr;
struct sockaddr_in us;
FILE	*tfp,*cfp;
#ifdef	MIPL_CACHE
GDBM_FILE dbf;
datum	data,key;
#endif
#ifdef	MIPL_SSL
X509*    server_cert;
char*    str;   
int	err;
#endif

static char *meths[] = { "GET","GET","POST","POST","PUT" } ;

static char *monnam[] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug",
	"Sep","Oct","Nov","Dec",0};

redirect:
	bign = bigindex = 0;
	usessl = 0;
#ifdef	MIPL_SSL
	ssl = 0;
#endif

	auth.last = auth.n;
	cp5 = 0;
	cp = b;
	j = 0;
	cname[0] = 0;
	if (strlen(url) >= MAXURL) *(url + MAXURL) = 0;
	   /* cant have a label in a secondary URL */
	if ((!mainflag) && (strchr(url,'#'))) return(NULL);
	if (*url == '#') /* if just a label */
	   {
		ourlabel[0] = 0;
		if (!*++url) return(NULL);
		strcpy(ourlabel,url);
		return(&dummyf);
	   }
	if (!strchr(url,':')) /* a partial URL */
	   {
		if (url[0] != '/') /* if entirely relative */
		   {
			strcpy(b,defstuff);
			strcat(b,url);
		   }
		else /* if just host (er whatevers) taken from default */
		   {
			if (strchr(defstuff,'/'))
			   {
				cpx = strchr(defstuff,'/');
				if (*(cpx + 1) == '/') cpx = strchr(cpx + 2,'/');
				i = cpx - defstuff;
				strncpy(b,defstuff,i);
				b[i] = 0;
				strcat(b,url);
				if (mainflag && strrchr(b,'/'))
				   {
					i = strrchr(b,'/') - b;
					strncpy(defstuff,b,i + 1);
					defstuff[i + 1] = 0;
				   }
			   } else return(NULL);
		   }
	   }
	else /* is a full URL */
	   {
		strcpy(b,url);
		if (mainflag && strrchr(b,'/'))
		   {
			i = strrchr(b,'/') - b;
			strncpy(defstuff,b,i + 1);
			defstuff[i + 1] = 0;
		   }
	   }
	cp1 = strchr(cp,':');
	if (cp1)
	   {
		cpx = cp1;
		*cp1 = 0;
		cp1 = cp; 
		cp = cpx + 1;
	   }
	cp2 = strstr(cp,"//");
	if (cp2)
	   {
		cpx = cp2;
		*cp2 = 0;
		cp = cp2 = cpx + 2;
		cp5 = strchr(cp,':');
		if (cp5)
		   {
			cpx = cp5;
			*cp5 = 0;
			cp5 = cp = cpx + 1;
		   }
	   }
	if (!strcmp(cp1,"xfer")) /* it's a transfer */
	   {
		mipl_xfer(cp);
		exit(0);
	   }
	if (!strcmp(cp1,"conf")) /* it's a conference test */
	   {
		conftest(cp);
		exit(0);
	   }
	if (!strcmp(cp1,"file"))  /* its a file URL */
	   {
		cp3 = 0;
		j = 1;
		cp4 = strchr(cp,'#');
		if (cp4)
		   {
			cpx = cp4;
			*cp4 = 0;
			cp4 = cp;
			cp = cpx + 1;
		   }
		else
		   {
			cp4 = cp;
			cp = 0;
		   }
 	   }
	if ((!strcmp(cp1,"http")) || (!strcmp(cp1,"https")))   /* it's http(s) */
	   {
		if (!strcmp(cp1,"https"))
		   {
#ifdef	MIPL_SSL
			if (!cp5) cp5 = "443";
			usessl = 1;
#else
			return(NULL);
#endif
		   }
		if (!cp5) cp5 = "80";
		j = 1;
		cp3 = strchr(cp,'/');
		if (cp3)
		   {
			cpx = cp3;
			*cp3 = 0;
			cp3 = cp;
			cp = cpx + 1;
		   }
		else
		   {
			if (cp2) 
			   {
				cp3 = cp; /* force slash even if not put */
				cp += strlen(cp);
			   }
		   }
		cp4 = strchr(cp,'#');
		if (cp4)
		   {
			cpx = cp4;
			*cp4 = 0;
			cp4 = cp;
			cp = cpx + 1;
		   }
		else
		   {
			cp4 = cp;
			cp = 0;
		   }
	   }
	if (!j) return(NULL); /* if unsupported type */

	/* okay we will now have:
		cp1 - service type (e.g. http)
		cp2 - domain address (e.g.foo.com)
		cp3 - binary flag (although it *is* a char *) signifying
			that a slash is to be inserted in front of the
		path contained in cp4.
		cp4 - file name path
		cp5 - tcp port number (e.g. 80)
		cp - label
	*/



/*
	printf("type:%s, domname:%s, port:%s, slash:%s, path:%s, label:%s\n",
		cp1,cp2,cp5,cp3,cp4,cp);
*/

again:
	s = -1;
	bign = bigindex = 0;	
	nocache = 0;
	cname[0] = 0;
	if (cp1 && ((!strcmp(cp1,"http")) || (!strcmp(cp1,"https"))))
	   {
		if (usessl)
		   {
			sprintf(cname,"https://%s:%s",cp2,cp5);
		   }
		else
		   {
			sprintf(cname,"http://%s:%s",cp2,cp5);
		   }
		if (cp3) strcat(cname,"/");
		if (cp4) strcat(cname,cp4);
	   } else nocache = 1;
	a[0] = 0;
	if (cp3) strcat(a,"/");
	strcat(a,cp4);
	if ((!nocache) && (method || strchr(a,'?'))) nocache = 1;
	cache_read = 0;
	cachehit = 0;
#ifdef	MIPL_CACHE
	if (!nocache)
	   {
		for(i = 0; i < CACHETRIES; i++)
		   {
			dbf = gdbm_open(CACHEDB,0,GDBM_READER,0,NULL);
			if (dbf) break;
		   }
		if (i >= CACHETRIES)
		   {
			mipl_error("Cant open DB!!!");
			nocache = 1;
			goto nocacheopen;
		   }
		key.dptr = cname;
		key.dsize = strlen(cname) + 1;
		data = gdbm_fetch(dbf,key);
		gdbm_close(dbf);
		if (data.dptr) /* if record found */
		   {
			strs1[0] = strs1[1] = strs1[2] = 0;
			finddelim(data.dptr,strs1,' ','\"');
			strcpy(buf,strs1[2]);
			cache_read = 1;
			if ((!mainflag) && (!mainchanged))
			   {
#ifdef	CACHEDEBUG
				printf("@@@@@@ main stable, using local cache\n");
#endif
				cachehit = 1; 
				goto cacheage;
			   }
		   }
	   }
#endif
nocacheopen:
	if (!strcmp(cp1,"file"))  /* its a file URL */
	   {
		fp = fopen(a,"r");
		if (!fp) return(NULL);
	   }
	else  /* it's an http request */
	   {
		printf("%%Looking up host... ");
		fflush(stdout);
		host = gethostbyname(cp2);
		if (!host)
		   {
			printf("Unable to find DNS entry for host %s\n",cp2);
			return(NULL);
		   }
		memcpy(&addr,*host->h_addr_list,sizeof(long));

		printf("Trying... ");
		fflush(stdout);
		s = socket(PF_INET,SOCK_STREAM,0);
		if (s == -1)
		   {
			perror("socket");
			killtemp();
			exit(255);
		   }
		us.sin_family = AF_INET;
		us.sin_port = htons(atoi(cp5));
		us.sin_addr.s_addr = addr;
	
		if (connect(s,(struct sockaddr *)&us,sizeof(us)) == -1)
		   {
			printf("%% unable to connect to host\n");
			return(NULL);
		   }
		printf("Open\n");
		fflush(stdout);
#ifdef	MIPL_SSL
		if (usessl)
		   {
			/* Now we have TCP conncetion. Start SSL negotiation. */
	
			ssl = SSL_new (ctx);
	                CHK_NULL(ssl);
			SSL_set_fd (ssl, s);
			err = SSL_connect (ssl);
	                CHK_SSL(err);

#ifdef	SSLDEBUG
			/* Following two steps are optional and not required for
			   data exchange to be successful. */               
			/* Get the cipher - opt */

			printf ("SSL connection using %s\n", SSL_get_cipher (ssl));

			/* Get server's certificate (note: beware of dynamic allocation) - opt */

			server_cert = SSL_get_peer_certificate (ssl);       CHK_NULL(server_cert);
			printf ("Server certificate:\n");

			str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
			CHK_NULL(str);
			printf ("\t subject: %s\n", str);
			Free (str);

			str = X509_NAME_oneline (X509_get_issuer_name  (server_cert),0,0);
			CHK_NULL(str);
			printf ("\t issuer: %s\n", str);
			Free (str); 

			/* We could do all sorts of certificate verification stuff here before
			   deallocating the certificate. */

			X509_free (server_cert);        
#endif
		   }
#endif
		j = 0;
		if ((method == GETMETHOD) || (method == POSTFORM))
		   {
			for(i = 0; i < form.npairs; i++)
			   {
				if (form.names[i]) j += strlen(form.names[i]) + 1;
				if (form.values[i]) j += strlen(form.values[i]);
			   }
			j += form.npairs - 1;
		   }
		tfp = 0;
#ifndef	NORECORD
		if ((method == POSTFILE) || (method == PUTFILE))
		   {
			tfp = fopen(rectemp,"r");
			if (!tfp)
			   {
				perror("rectemp post/put open");
				killtemp();
				exit(255);
			   }
			if (fseek(tfp,0L,SEEK_END) == -1)
			   {
				perror("rectemp post/put fseek");
				killtemp();
				exit(255);
			   }
			j = ftell(tfp);
			rewind(tfp);
		   }
#endif
		cpx = malloc(j + HEADERFUDGE);
		if (!cpx)
		   {
			perror("malloc");
			killtemp();
			exit(255);
		   }
		sprintf(cpx,"%s %s",meths[method],a);
		if (method == GETMETHOD)
		   {
			strcat(cpx,"?");
			for(i = 0,k = 0; i < form.npairs; i++)
			   {
				if (!form.names[i]) continue;
				if (k) strcat(cpx,"&");
				 sprintf(cpx + strlen(cpx),"%s=",form.names[i]);
				if (form.values[i]) strcat(cpx,form.values[i]);
				k = 1;
			   }
		   }
		strcat(cpx," HTTP/1.0\r\n");
		if (myhostname)
			sprintf(cpx + strlen(cpx),"Host: %s\r\n",myhostname);
		sprintf(cpx + strlen(cpx),"Accept:\
 sun-deskset-message, mail-file, default, postscript-file, audio-file,\
 x-sun-attachment, text/richtext, text/enriched, application/andrew-inset,\
 x-be2, message/delivery-status, message/external-body, message/partial,\
 application/pgp, application/postscript, image/gif, image/*, video/*,\
 audio/x-wav, audio/basic, audio/wav, audio/x-wav, audio/au, audio/basic,\
 text/html, text/plain, application/x-wais-source, application/html,\
 */*;q=0.001\r\nAccept-Encoding: gzip, compress\r\nAccept-Language: en\r\n\
User-Agent: lamb/%s\r\n",MIPL_VERSION);
		
		if (auth.last < auth.n)
		   {
			memset(a,0,sizeof(a));
			sprintf(a,"%s:%s",auth.user[auth.last],auth.pswd[auth.last]);
			strcat(cpx,"Authorization: Basic ");
			HTUU_encode(a,strlen(a),cpx + strlen(cpx));
			strcat(cpx,"\r\n");
		   }

		cl = strlen(cpx);
		if (method == POSTFORM)
		   {
			sprintf(cpx + strlen(cpx),"\
Pragma: no-cache\r\nCache-Control: no-cache\r\n\
Content-type: application/x-www-form-urlencoded\r\nContent-length: %d\r\n\r\n",j);
			for(i = 0,k = 0; i < form.npairs; i++)
			   {
				if (!form.names[i]) continue;
				if (k) strcat(cpx,"&");
				 sprintf(cpx + strlen(cpx),"%s=",form.names[i]);
				if (form.values[i]) strcat(cpx,form.values[i]);
				k = 1;
			   }
			cl = strlen(cpx);
		   }

#ifndef	NORECORD
		if ((method == POSTFILE) || (method == PUTFILE))
		   {
			sprintf(cpx + strlen(cpx),"\
Pragma: no-cache\r\nCache-Control: no-cache\r\n\
Content-type: application/octet-stream\r\nContent-length: %d\r\n\r\n",j);
			cl = strlen(cpx);
			for(i = 0; i < j; i++)
			   {
				c = fgetc(tfp);
				if (c == EOF)
				   {
					perror("Error getting from rectemp\n");
					killtemp();
					exit(255);
				   }
				cpx[cl + i] = c;
			   }
			cl += j;
			fclose(tfp);
		   }
#endif

		strcpy(cpx + cl,"\r\n");			
		cl += 2;

#ifdef	MIPL_SSL
		if (usessl)
		   {
			  err = SSL_write (ssl, cpx, cl);
			  CHK_SSL(err);
			  shutdown (s, 1);  /* Half close, send EOF to server. */  
		   }
		else
#endif
		for(i = 0; i < cl; i += j)
		   {
			j = cl;
			if (j > 1024) j = 1024;
			if (send(s,cpx + i,j,0) != j)
			   {
				mipl_error("unable to send request to host");
				perror("send");
				if (s != -1) close(s);
#ifdef	MIPL_SSL
				if (ssl) SSL_free(ssl);
#endif
				return(NULL);
			   }
		   }
		free(cpx);
		clen = -1;
		cmod[0] = 0;		
		for(j = 0,k = 0;;)
		   {
			i = strin(s,a,sizeof(a) - 1);
			if (i == -1)
			   {
				mipl_error("unable to receive header from host");
				if (s != -1) close(s);
#ifdef	MIPL_SSL
				if (ssl) SSL_free(ssl);
#endif
				return(NULL);
			   }
			if (!a[0]) break;
			if (j == 0)
			   {
				cpx = strchr(a,' ');
				if (!cpx)
				   {
					mipl_error("malformed header");
#ifdef	MIPL_SSL
					if (ssl) SSL_free(ssl);
#endif
					if (s != -1) close(s);
					return(NULL);
				   }
				*cpx = 0;				
				k = atoi(cpx + 1);  /* get status */
				j++;
				continue;				
			   }
			cpx = strchr(a,':');
			if (!cpx) continue;
			*cpx++ = 0;
			while(*cpx && (*cpx == ' ')) cpx++;			
			if (!strcmp(a,"Location"))
			   {
				strcpy(a,cpx);
				if (s != -1) close(s);
#ifdef	MIPL_SSL
				if (ssl) SSL_free(ssl);
#endif
				url = a;
				printf("%Redirecting...\n");
				goto redirect;
			   }
			if ((!strcmp(a,"Content-Length")) ||
			 (!strcmp(a,"Content-length")))
			   {
				clen = atoi(cpx);
				continue;
			   }
			if ((!strcmp(a,"Pragma")) ||
			 (!strcmp(a,"Cache-Control")))
			   {
				if (strstr(cpx,"no-cache"))
				   {
					nocache = 1;
				   }
				continue;
			   }
			if ((!strcmp(a,"Last-Modified")) ||
			  (!strcmp(a,"Last-modified")))
			   {
				for(i = 0; cpx[i]; i++)
					if (cpx[i] == ':') cpx[i] = ' ';

				i = finddelim(cpx,strs,' ','\"');
				if (i < 8) continue;
				for(i = 0; monnam[i]; i++)
				   {
					if (!strcmp(strs[2],monnam[i])) break;
				   }
				if (!monnam[i]) continue;
				sprintf(cmod,"%4.4s%02d%2.2s%2.2s%2.2s%2.2s",strs[3],i + 1,
					strs[1],strs[4],strs[5],strs[6]);
			   }
			if ((!strcmp(a,"WWW-Authenticate")) ||
			  (!strcmp(a,"WWW-authenticate")))
			   {
				if (s != -1) close(s);
#ifdef	MIPL_SSL
				if (ssl) SSL_free(ssl);
#endif
				cpy = strchr(cpx,'\"');
				if (cpy) cpx = cpy + 1;
				if (cpx[strlen(cpx) - 1] == '\"')
					cpx[strlen(cpx) - 1] = 0;
				if ((auth.last >= auth.n) ||
				   (strcmp(auth.realm[auth.last],cpx)))
				   {
					 for(i = 0; i < auth.n; i++)
					   {
						if (!strcmp(auth.realm[i],cpx))
						   {
							auth.last = i;
							goto again;
						   }
					   }
				   }
				if (auth.last >= auth.n)
				   {
					  /* allocate a new one */
					auth.last = auth.n++;
					cpy = malloc(strlen(cpx) + 1);
					if (!cpy)
					   {
						perror("malloc");
						killtemp();
						exit(255);
					   }
					strcpy(cpy,cpx);
					auth.realm[auth.last] = cpy;
				   }
				mipl_flush(MIPL_READ);
				if (!mipl_askauth(cpx,a,buf)) return(NULL);
				cpy = malloc(strlen(a) + 1);
				if (!cpy)
				   {
					perror("malloc");
					killtemp();
					exit(255);
				   }
				strcpy(cpy,a);
				if (auth.user[auth.last]) free(auth.user[auth.last]);
				auth.user[auth.last] = cpy;
				cpy = malloc(strlen(buf) + 1);
				if (!cpy)
				   {
					perror("malloc");
					killtemp();
					exit(255);
				   }
				strcpy(cpy,buf);
				if (auth.pswd[auth.last]) free(auth.pswd[auth.last]);
				auth.pswd[auth.last] = cpy;
				goto again;
			   }			
#ifdef	DUMPRX 
			printf("Got header: %s\n",a);
#endif 
		   }

		if ((k < 200) || (k > 399)) 
		   {
			sprintf(outbuf,"Server Error %d getting script %s\n",k,url);
			mipl_error(outbuf);
			if (s != -1) close(s);
#ifdef	MIPL_SSL
			if (ssl) SSL_free(ssl);
#endif
			return(NULL);
		   }			
#ifdef	MIPL_CACHE
		cachehit = 0;
		if (cache_read)
		   {
			if ((!strcmp(strs1[0],cmod)) && (atoi(strs1[1]) == clen))
			   {
				cachehit = 1;
			   }
			free(data.dptr);
		   }
cacheage:
		if (mainflag) mainchanged = nocache || (!cachehit);
		if ((!nocache) && cachehit)
		   {
			cfp = fopen(buf,"r");
			if (!cfp) 
			   {
				perror("cache file open for read");
				nocache = 1;
				cachehit = 0;
			   }
		   }
#endif
#ifdef	NORECORD
		tfp = tmpfile();
#else
		unlink(rectemp);
		tfp = fopen(rectemp,"w+");
#endif
		if (!tfp)
		   {
			perror("temp file open");
			killtemp();
			exit(255);
		   }
		if ((!cachehit) || nocache)  /* if not in cache */
		   {
#ifdef	MIPL_CACHE
			if (!nocache)
			   {
				cpt = tempnam(CACHEDIR,CACHETMP);
				if (!cpt)
				   {
					perror("Cant get temp file name!!");
					if (s != -1) close(s);
#ifdef	MIPL_SSL
					if (ssl) SSL_free(ssl);
#endif
					return(NULL);
				   }
				strcpy(buf,cpt);
#ifdef	CACHEDEBUG
				printf("@@@@@@ putting %s into %s\n",cname,buf);
#endif
				cfp = fopen(cpt,"w");
				if (!cfp)
				   {
					perror("Cant open cache file for write!!!");
					if (s != -1) close(s);
#ifdef	MIPL_SSL
					if (ssl) SSL_free(ssl);
#endif
					return(NULL);
				   }
			   }
#endif
			for(;;)
			   {
				if (!mainflag)
				   {
					if (bigindex < bign)
					   {
						i = bign - bigindex;
						if (i > (sizeof(a) - 1))
							i = sizeof(a) - 1;
						bcopy(bigbuf + bigindex,a,i);
						bigindex += i;
					   }
					else
					   {
#ifdef	MIPL_SSL
						if (usessl)
						   {
							i = SSL_read(ssl,a,sizeof(a) - 1);
						   }
						else
#endif
						i = recv(s,a,sizeof(a) - 1,0);
					   }
					if (i < 1) break;
					if (fwrite(a,1,i,tfp) != i)
					   {
						perror("temp file write");
						killtemp();
						exit(255);
					   }
#ifdef	MIPL_CACHE
					if (nocache) continue;
					if (fwrite(a,1,i,cfp) != i)
					   {
						perror("cache file write");
						killtemp();
						exit(255);
					   }
#endif
					continue;
				   }
				i = strin(s,a,sizeof(a) - 1);
				if (i == -1) break;
				strcat(a,"\n");
#ifdef	DUMPRX
				printf("Got file line: %s\n",a);
#endif
				if (fwrite(a,1,strlen(a),tfp) != strlen(a))
				   {
					perror("temp file write");
					killtemp();
					exit(255);
				   }
				if (nocache) continue;
#ifdef	MIPL_CACHE
				if (fwrite(a,1,strlen(a),cfp) != strlen(a))
				   {
					perror("cache file write");
					killtemp();
					exit(255);
				   }
#endif

			   }
#ifdef	MIPL_CACHE
			if (!nocache)
			   {
				fclose(cfp);
				for(i = 0; i < CACHETRIES; i++)
				   {
					dbf = gdbm_open(CACHEDB,0,GDBM_WRITER,0,NULL);
					if (dbf) break;
				   }
				if (i >= CACHETRIES)
				   {
					mipl_error("Cant open DB (write)!!!");
					nocache = 1;
				   }
				else
				   {
					sprintf(a,"%s %d %s",cmod,clen,buf);
					data.dptr = a;
					data.dsize = strlen(a) + 1;
					gdbm_store(dbf,key,data,GDBM_REPLACE);
					gdbm_close(dbf);
				   }
			   }
#endif
			if (s != -1) close(s);
#ifdef	MIPL_SSL
			if (ssl) SSL_free(ssl);
#endif
		   }
#ifdef	MIPL_CACHE
		else /* in cache */
		   {
#ifdef	CACHEDEBUG
			printf("@@@@@@ getting %s from %s\n",cname,buf);
#endif
			if (s != -1) close(s);  /* close server conn already */
#ifdef	MIPL_SSL
			if (ssl) SSL_free(ssl);
#endif
			  /* copy from cache file */
			while((i = fread(a,1,sizeof(a),cfp)) > 0)
			   {
				if (fwrite(a,1,i,tfp) != i)
				   {
					perror("temp file write");
					killtemp();
					exit(255);
				   }
			   }
			fclose(cfp);
		   }
#endif
		rewind(tfp);
	   }


	if (mainflag)
	   {
		ourlabel[0] = 0;
		if (cp)
		   {
			if (strlen(cp) >= MAXURL) *(cp + MAXURL) = 0;
			strcpy(ourlabel,cp);
		   }
	   }
	return(tfp);
}

int openurl(url,method)
char *url;
int method;
{
int	c,i;
FILE	*tfp;

	tfp = readurl(url,method,1);
	if (!tfp) return(-1);
	if (tfp == &dummyf) return(0);
	fp = tfp;
	freetag(&maintag);
	memset(&maintag,0,sizeof(TAG));
	for(i = 0;;)
	   {
		c = mip_cin();
		if (c == EOF) 
		   {
			killtemp();
			return(-1);
		   }
		switch(i)
		   {
		    case 0:
			if (c != '<') continue;
			i = 1;
			break;
		    case 1:
			if (isspace(c)) continue;
			if ((c != 'M') && (c != 'm')) i = 0; else i = 2;
			break;
		    case 2:					
			if ((c != 'I') && (c != 'i')) i = 0; else i = 3;
			break;
		    case 3:					
			if ((c != 'P') && (c != 'p')) i = 0; else i = 4;
			break;
		    case 4:					
			if ((c != 'L') && (c != 'l')) i = 0; else i = 5;
			break;
		    case 5:					
			if (isspace(c)) continue;
			if (c != '>') i = 0; else i = 6;
			break;
		   }
		if (i == 6) break;
	   }
	  /* start the machine.. call it again if theres a comment */
	do i = gyrate(&maintag); while (i == 1);
	if (i == -1) return(-1);
	fclose(fp);
	return(0);
}

int mip_cin()
{

	return(fgetc(fp));
}

strin(s,str,max)
int s;
char *str;
int max;
{
int	togo,i;
char	c;

	togo = -1;
	for(;;)
	   {
		if (bigindex >= bign) /* if to suck another chunk */
		   {
			fflush(stdout);
#ifdef	MIPL_SSL
			if (usessl)
			   {
				if (bign && (bign < sizeof(bigbuf))) return(-1);
				i = SSL_read(ssl,bigbuf,sizeof(bigbuf));
			   }
			else
#endif
			i = recv(s,bigbuf,sizeof(bigbuf),0);
			if (i < 1)
			   {
				return(-1);
			   }
			bigindex = 0;
			bign = i;
		   }
		c = bigbuf[bigindex++];		
		if (c == 13) continue;
		if (c == 10)
		   {
			*str = 0;
			return(0);
		   }
		*str++ = c;
		if (!max--)
		   {
			*str = 0;
			return(0);
		   }
	   }
}


/* MODULE							HTUU.c
**			UUENCODE AND UUDECODE
**
** ACKNOWLEDGEMENT:
**	This code is taken from rpem distribution, and was originally
**	written by Mark Riordan.
**
** AUTHORS:
**	MR	Mark Riordan	riordanmr@clvax1.cl.msu.edu
**	AL	Ari Luotonen	luotonen@dxcern.cern.ch
**
** HISTORY:
**	Added as part of the WWW library and edited to conform
**	with the WWW project coding standards by:	AL  5 Aug 1993
**	Originally written by:				MR 12 Aug 1990
**	Original header text:
** -------------------------------------------------------------
**  File containing routines to convert a buffer
**  of bytes to/from RFC 1113 printable encoding format.
**
**  This technique is similar to the familiar Unix uuencode
**  format in that it maps 6 binary bits to one ASCII
**  character (or more aptly, 3 binary bytes to 4 ASCII
**  characters).  However, RFC 1113 does not use the same
**  mapping to printable characters as uuencode.
**
**  Mark Riordan   12 August 1990 and 17 Feb 1991.
**  This code is hereby placed in the public domain.
** -------------------------------------------------------------
**
** BUGS:
**
**
*/

static char six2pr[64] = {
    'A','B','C','D','E','F','G','H','I','J','K','L','M',
    'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
    'a','b','c','d','e','f','g','h','i','j','k','l','m',
    'n','o','p','q','r','s','t','u','v','w','x','y','z',
    '0','1','2','3','4','5','6','7','8','9','+','/'
};

static unsigned char pr2six[256];


/*--- function HTUU_encode -----------------------------------------------
 *
 *   Encode a single line of binary data to a standard format that
 *   uses only printing ASCII characters (but takes up 33% more bytes).
 *
 *    Entry    bufin    points to a buffer of bytes.  If nbytes is not
 *                      a multiple of three, then the byte just beyond
 *                      the last byte in the buffer must be 0.
 *             nbytes   is the number of bytes in that buffer.
 *                      This cannot be more than 48.
 *             bufcoded points to an output buffer.  Be sure that this
 *                      can hold at least 1 + (4*nbytes)/3 characters.
 *
 *    Exit     bufcoded contains the coded line.  The first 4*nbytes/3 bytes
 *                      contain printing ASCII characters representing
 *                      those binary bytes. This may include one or
 *                      two '=' characters used as padding at the end.
 *                      The last byte is a zero byte.
 *             Returns the number of ASCII characters in "bufcoded".
 */

int HTUU_encode(bufin,nbytes,bufcoded)
unsigned char * bufin;
unsigned int nbytes;
char *bufcoded;
{
/* ENC is the basic 1 character encoding function to make a char printing */
#define ENC(c) six2pr[c]

   register char *outptr = bufcoded;
   unsigned int i;
   /* This doesn't seem to be needed (AL):   register unsigned char *inptr  = bufin; */

   for (i=0; i<nbytes; i += 3) {
      *(outptr++) = ENC(*bufin >> 2);            /* c1 */
      *(outptr++) = ENC(((*bufin << 4) & 060) | ((bufin[1] >> 4) & 017)); /*c2*/
      *(outptr++) = ENC(((bufin[1] << 2) & 074) | ((bufin[2] >> 6) & 03));/*c3*/
      *(outptr++) = ENC(bufin[2] & 077);         /* c4 */

      bufin += 3;
   }

   /* If nbytes was not a multiple of 3, then we have encoded too
    * many characters.  Adjust appropriately.
    */
   if(i == nbytes+1) {
      /* There were only 2 bytes in that last group */
      outptr[-1] = '=';
   } else if(i == nbytes+2) {
      /* There was only 1 byte in that last group */
      outptr[-1] = '=';
      outptr[-2] = '=';
   }
   *outptr = '\0';
   return(outptr - bufcoded);
}


/*--- function HTUU_decode ------------------------------------------------
 *
 *  Decode an ASCII-encoded buffer back to its original binary form.
 *
 *    Entry    bufcoded    points to a uuencoded string.  It is 
 *                         terminated by any character not in
 *                         the printable character table six2pr, but
 *                         leading whitespace is stripped.
 *             bufplain    points to the output buffer; must be big
 *                         enough to hold the decoded string (generally
 *                         shorter than the encoded string) plus
 *                         as many as two extra bytes used during
 *                         the decoding process.
 *             outbufsize  is the maximum number of bytes that
 *                         can fit in bufplain.
 *
 *    Exit     Returns the number of binary bytes decoded.
 *             bufplain    contains these bytes.
 */
int HTUU_decode(bufcoded,bufplain,outbufsize)
char *bufcoded;
unsigned char *bufplain;
int outbufsize;
{
/* single character decode */
#define DEC(c) pr2six[c]
#define MAXVAL 63

   static int first = 1;

   int nbytesdecoded, j;
   register char *bufin = bufcoded;
   register unsigned char *bufout = bufplain;
   register int nprbytes;

   /* If this is the first call, initialize the mapping table.
    * This code should work even on non-ASCII machines.
    */
   if(first) {
      first = 0;
      for(j=0; j<256; j++) pr2six[j] = MAXVAL+1;

      for(j=0; j<64; j++) pr2six[(unsigned char)six2pr[j]] = (unsigned char)j;
#if 0
      pr2six['A']= 0; pr2six['B']= 1; pr2six['C']= 2; pr2six['D']= 3; 
      pr2six['E']= 4; pr2six['F']= 5; pr2six['G']= 6; pr2six['H']= 7; 
      pr2six['I']= 8; pr2six['J']= 9; pr2six['K']=10; pr2six['L']=11; 
      pr2six['M']=12; pr2six['N']=13; pr2six['O']=14; pr2six['P']=15; 
      pr2six['Q']=16; pr2six['R']=17; pr2six['S']=18; pr2six['T']=19; 
      pr2six['U']=20; pr2six['V']=21; pr2six['W']=22; pr2six['X']=23; 
      pr2six['Y']=24; pr2six['Z']=25; pr2six['a']=26; pr2six['b']=27; 
      pr2six['c']=28; pr2six['d']=29; pr2six['e']=30; pr2six['f']=31; 
      pr2six['g']=32; pr2six['h']=33; pr2six['i']=34; pr2six['j']=35; 
      pr2six['k']=36; pr2six['l']=37; pr2six['m']=38; pr2six['n']=39; 
      pr2six['o']=40; pr2six['p']=41; pr2six['q']=42; pr2six['r']=43; 
      pr2six['s']=44; pr2six['t']=45; pr2six['u']=46; pr2six['v']=47; 
      pr2six['w']=48; pr2six['x']=49; pr2six['y']=50; pr2six['z']=51; 
      pr2six['0']=52; pr2six['1']=53; pr2six['2']=54; pr2six['3']=55; 
      pr2six['4']=56; pr2six['5']=57; pr2six['6']=58; pr2six['7']=59; 
      pr2six['8']=60; pr2six['9']=61; pr2six['+']=62; pr2six['/']=63;
#endif
   }

   /* Strip leading whitespace. */

   while(*bufcoded==' ' || *bufcoded == '\t') bufcoded++;

   /* Figure out how many characters are in the input buffer.
    * If this would decode into more bytes than would fit into
    * the output buffer, adjust the number of input bytes downwards.
    */
   bufin = bufcoded;
   while(pr2six[(unsigned char)*(bufin++)] <= MAXVAL);
   nprbytes = bufin - bufcoded - 1;
   nbytesdecoded = ((nprbytes+3)/4) * 3;
   if(nbytesdecoded > outbufsize) {
      nprbytes = (outbufsize*4)/3;
   }

   bufin = bufcoded;
   
   while (nprbytes > 0) {
      *(bufout++) = (unsigned char) (DEC(*bufin) << 2 | DEC(bufin[1]) >> 4);
      *(bufout++) = (unsigned char) (DEC(bufin[1]) << 4 | DEC(bufin[2]) >> 2);
      *(bufout++) = (unsigned char) (DEC(bufin[2]) << 6 | DEC(bufin[3]));
      bufin += 4;
      nprbytes -= 4;
   }
   
   if(nprbytes & 03) {
      if(pr2six[bufin[-2]] > MAXVAL) {
         nbytesdecoded -= 2;
      } else {
         nbytesdecoded -= 1;
      }
   }

   return(nbytesdecoded);
}


