/*
 * Configuration and startup program for:
 *
 * BSD Telephony Of Mexico "Tormenta" card LINUX driver, version 1.5 1/1/01
 * 
 * Working with the "Tormenta ISA" Card 
 * 
 *
 *   Part of the "Zapata" Computer Telephony Technology.
 *
 *   See http://www.bsdtelephony.com.mx
 *
 *
 *  The technologies, software, hardware, designs, drawings, scheumatics, board
 *  layouts and/or artwork, concepts, methodologies (including the use of all
 *  of these, and that which is derived from the use of all of these), all other
 *  intellectual properties contained herein, and all intellectual property
 *  rights have been and shall continue to be expressly for the benefit of all
 *  mankind, and are perpetually placed in the public domain, and may be used,
 *  copied, and/or modified by anyone, in any manner, for any legal purpose,
 *  without restriction.
 *
 */

#include <stdio.h>
#include <tor.h>
#include <fcntl.h>

main(int argc, char *argv[])
{
char *fname = "/etc/tor.conf";  /* default config file name */
char	buf[100];  /* file line buffer */
int	i,j,k,f,lineno; /* some ints */
FILE	*fp;  /* config file pointer */
struct tor_lineconfig l;  /* line config structure */
struct tor_chanconfig c;  /* channel config structure */

	if (argc > 1) fname = argv[1];
	fp = fopen(fname,"r");  /* open config file */
	if (!fp) /* if cannot open, bail */
	   {
		fprintf(stderr,"Cannot open config file %s\n",fname);
		perror("config file open");
		exit(255);
	   }		
	f = open("/dev/tor0",O_RDWR); /* open master tormenta device */
	if (f == -1) /* if can not open */
	   {
		perror("/dev/tor0 open");
		exit(255);
	   }
	bzero(&l,sizeof(l)); /* zero out the lineconfig structure */
	lineno = 1; /* start with file line 1 */
	while(fgets(buf,sizeof(buf) - 1,fp)) /* get next line */
	   {
		if (!buf[0]) continue; /* skip completely blank line */
		  /* chop at comment, if any */
		for(i = 0; buf[i]; i++) if (buf[i] == '#') buf[i] = 0;
		if (!buf[0]) continue; /* skip blank line */
		if (buf[0] == '#') continue; /* skip comment */
		switch(lineno)
		   {
		    case 1: /* line 1, first span */
 			if (sscanf(buf,"%i %i %i %i",&l.lbo[0],&i,&j,&k) < 4)
			   {
				fprintf(stderr,"file line %d invalid!!\n",lineno);
				exit(255);
			   }
			l.lineconfig[0] = 0;
			if (i) l.lineconfig[0] |= TOR_CONFIG_ESF;
			if (j) l.lineconfig[0] |= TOR_CONFIG_B8ZS;
			if (k) l.lineconfig[0] |= TOR_CONFIG_NOTOPEN;
			break;
		    case 2: /* line 2, second span */
 			if (sscanf(buf,"%i %i %i %i",&l.lbo[1],&i,&j,&k) < 4)
			   {
				fprintf(stderr,"file line %d invalid!!\n",lineno);
				exit(255);
			   }
			l.lineconfig[1] = 0;
			if (i) l.lineconfig[1] |= TOR_CONFIG_ESF;
			if (j) l.lineconfig[1] |= TOR_CONFIG_B8ZS;
			if (k) l.lineconfig[1] |= TOR_CONFIG_NOTOPEN;
			break;
		    case 3: /* line 3, sync sources */
 			if (sscanf(buf,"%i %i",&l.psync,&l.ssync) < 2)
			   {
				fprintf(stderr,"file line %d invalid!!\n",lineno);
				exit(255);
			   }
			break;
		   default: /* next lines are channels */
 			if (sscanf(buf,"%i %i",&c.chan,&c.sigtype) < 2)
			   {
				fprintf(stderr,"file line %d invalid!!\n",lineno);
				exit(255);
			   }
			if (ioctl(f,TOR_CHANCONFIG,&c) == -1)
			   {
				fprintf(stderr,"Unable to configure for channel %d\n",c.chan);
				perror("chanconfig IOCTL");
				exit(255);
			   }
			break;
		   }
		lineno++;
	   }
	if (lineno < 4)
	   {
		fprintf(stderr,"File line(s) missing!!\n");
		exit(255);
	   }
	if (ioctl(f,TOR_STARTUP,&l) == -1)
	   {
		perror("startup IOCTL");
		exit(255);
	   }
	exit(0);
}
